from dataclasses import dataclass

import numpy as np

from ssip.basic_functions import (
    BinMatrix,
    CSScode,
    column_echelon_form,
    find_paired_basis,
    find_X_basis,
    find_Z_basis,
    image_basis_calc,
    kernel_basis_calc,
    multiply_F2,
    vec_addition,
)


@dataclass
class MergeResult:
    """A collection of data which informs the user about a merge, or
    multiple merges, which were performed.

    Attributes:
        Code: The output merged code.
        MergeMap: The map on qubits from the input to output code.
        NewZStabs: The indices of any new Z stabiliser generators.
        NewXStabs: The indices of any new X stabiliser generators.
        NewQubits: The indices of any new data qubits.
        NewZLogicals: A basis for the space of new Z logicals
        introduced by the merge, if there are any.
        NewXLogicals: A basis for the space of new X logicals
        introduced by the merge, if there are any.
        OldZLogicals: A basis for the space of old Z logicals
        retained by the merge, if there are any.
        OldXLogicals: A basis for the space of old X logicals
        retained by the merge, if there are any.

    """

    Code: CSScode
    # if the merge was a Z-merge, this is the chain map at degree 1;
    # if X-merge, cochain map at degree 1
    MergeMap: BinMatrix
    NewZStabs: list
    NewXStabs: list
    NewQubits: list
    NewZLogicals: list
    NewXLogicals: list
    OldZLogicals: list
    OldXLogicals: list


def find_logical_splitting(
    old_code: CSScode,
    new_code: CSScode,
    merge_map: BinMatrix,
    num_before_logicals: int,
    num_after_logicals: int,
    num_merges: int = 1,
    basis: str = "Z",
):
    if basis != "Z" and basis != "X":
        raise ValueError("Must enter Z or X basis.")
    num_new_logicals = num_after_logicals - num_before_logicals + num_merges
    hom_before = None
    if basis == "Z":
        hom_before = np.array(find_Z_basis(old_code)).T
    else:
        hom_before = np.array(find_X_basis(old_code)).T
    if hom_before.shape[1] != num_before_logicals:
        raise RuntimeError("Dimensions do not match up when returning merge data.")
    hom_before_in_merged_code = multiply_F2(merge_map, hom_before)

    im_basis_after = None
    if basis == "Z":
        im_basis_after = np.array(image_basis_calc(new_code.PZ.T))
    else:
        im_basis_after = np.array(image_basis_calc(new_code.PX.T))
    Q = np.array(kernel_basis_calc(im_basis_after))

    ker_before_in_after_hom = multiply_F2(Q, hom_before_in_merged_code)

    mat1 = column_echelon_form(ker_before_in_after_hom)

    num_old_logicals = num_before_logicals - num_merges
    vecs_to_add = [0] * num_old_logicals
    height = ker_before_in_after_hom.shape[0]
    for i in np.arange(num_old_logicals):
        v = [0] * (mat1.shape[0] - height)
        for j in np.arange(height, mat1.shape[0]):
            v[j - height] = mat1[j][i]
        vecs_to_add[i] = v

    logical_basis = [0] * num_old_logicals
    for i in np.arange(num_old_logicals):
        vecs = [
            list(hom_before_in_merged_code[:, j])
            for j in np.arange(len(vecs_to_add[i]))
            if vecs_to_add[i][j]
        ]

        logical_basis[i] = vec_addition(vecs)

    S = np.array(kernel_basis_calc(ker_before_in_after_hom.T))

    if basis == "Z":
        ker_after = np.array(kernel_basis_calc(new_code.PX)).T
    else:
        ker_after = np.array(kernel_basis_calc(new_code.PZ)).T
    ker_after_quotient_old_logicals = multiply_F2(S, multiply_F2(Q, ker_after))
    mat = column_echelon_form(ker_after_quotient_old_logicals, True)

    height = ker_after_quotient_old_logicals.shape[0]
    vecs_to_add = [0] * num_new_logicals
    for i in np.arange(num_new_logicals):
        v = [0] * (mat.shape[0] - height)
        for j in np.arange(height, mat.shape[0]):
            v[j - height] = mat[j][i]
        vecs_to_add[i] = v

    gauge_basis = [0] * num_new_logicals
    for i in np.arange(num_new_logicals):
        vecs = [
            list(ker_after[:, j])
            for j in np.arange(len(vecs_to_add[i]))
            if vecs_to_add[i][j]
        ]
        gauge_basis[i] = vec_addition(vecs)

    if basis == "Z":
        new_Z_logicals = gauge_basis
        old_Z_logicals = logical_basis
        total_Z_basis = gauge_basis + logical_basis

        trial_X_basis = find_X_basis(new_code)
        total_X_basis = find_paired_basis(total_Z_basis, trial_X_basis)
        new_X_logicals = total_X_basis[:num_new_logicals]
        old_X_logicals = total_X_basis[num_new_logicals:]

        return (new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals)
    else:
        new_X_logicals = gauge_basis
        old_X_logicals = logical_basis
        total_X_basis = gauge_basis + logical_basis

        trial_Z_basis = find_Z_basis(new_code)
        total_Z_basis = find_paired_basis(total_X_basis, trial_Z_basis)
        new_Z_logicals = total_Z_basis[:num_new_logicals]
        old_Z_logicals = total_Z_basis[num_new_logicals:]

        return (new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals)



# Functions to generate tensor & lifted product CSS codes, following https://arxiv.org/abs/2012.04068
import numpy as np

from ssip.basic_functions import BinMatrix
from ssip.code_examples import CSScode


def tensor_product(A: BinMatrix, B: BinMatrix) -> CSScode:
    """Calculate the tensor product in the category of chain complexes
    given two length 1 chain complexes, i.e. classical codes. This produces
    a lengeth 2 chain complex, i.e. CSS code.

    Args:
        A: The first classical code.
        B: The second classical code.

    Return:
        The tensor product qubit CSS code.
    """
    del2a = np.kron(A, np.eye(B.shape[1]))
    del2b = np.kron(np.eye(A.shape[1]), B)
    del2 = np.vstack((del2a, del2b))

    del1a = np.kron(np.eye(A.shape[0]), B)
    del1b = np.kron(A, np.eye(B.shape[0]))
    del1 = np.hstack((del1a, del1b))

    return CSScode(del2.T, del1)


def lift_connected_surface_codes(l: int, L: int) -> CSScode:
    """Construct the lift-connected surface code with base code
    paramaterised by l and lift matrix parameterised by L. See
    p6, https://arxiv.org/abs/2401.02911. Assumes L > 1.

    Args:
        l: Parameter for base code.
        L: Parameter for lift matrix.

    Return:
        The lift-connected surface code LCS(l, L).
    """

    H_rep = np.zeros((l, l + 1))
    H_int = np.zeros((l, l + 1))
    for i in np.arange(l):
        H_rep[i][i] = 1
        H_rep[i][i + 1] = 1
        H_int[i][i + 1] = 1

    identity_L = np.eye(L)
    permutation = np.roll(identity_L, 1, axis=1)

    identity_l = np.eye(l)
    identity_l1 = np.eye(l + 1)

    P_X_rep = np.kron(
        np.hstack((np.kron(identity_l1, H_rep), np.kron(H_rep.T, identity_l))),
        identity_L,
    )
    P_X_int = np.hstack(
        (
            np.kron(np.kron(identity_l1, H_int), permutation),
            np.kron(np.kron(H_int.T, identity_l), permutation.T),
        )
    )
    P_X = P_X_rep + P_X_int

    P_Z_rep = np.kron(
        np.hstack((np.kron(H_rep, identity_l1), np.kron(identity_l, H_rep.T))),
        identity_L,
    )
    P_Z_int = np.hstack(
        (
            np.kron(np.kron(H_int, identity_l1), permutation),
            np.kron(np.kron(identity_l, H_int.T), permutation.T),
        )
    )
    P_Z = P_Z_rep + P_Z_int

    return CSScode(P_Z, P_X)


def bivariate_bicycle_code(
    l: int, m: int, powers_A: tuple[list[int]], powers_B: tuple[list[int]]
) -> CSScode:
    """Construct the bivariate bicycle code with circulants of size l and m
    respectively. See p6, https://arxiv.org/abs/2401.02911.

    Args:
        l: Size of first circulants in the tensor product.
        m: Size of second circulants in the tensor product.
        powers_A: Powers of x and y representing first circulant matrix A.
        powers_B: Powers of x and y representing second circulant matrix B.

    Return:
        The bivariate bicycle code.
    """

    identity_l = np.eye(l)
    permutation_l = np.roll(identity_l, 1, axis=1)
    identity_m = np.eye(m)
    permutation_m = np.roll(identity_m, 1, axis=1)
    x = np.kron(permutation_l, identity_m)
    y = np.kron(identity_l, permutation_m)

    A = np.zeros((x.shape[0], x.shape[1]))
    for i in powers_A[0]:
        A += np.linalg.matrix_power(x, i)
    for i in powers_A[1]:
        A += np.linalg.matrix_power(y, i)

    B = np.zeros((x.shape[0], x.shape[1]))
    for i in powers_B[0]:
        B += np.linalg.matrix_power(x, i)
    for i in powers_B[1]:
        B += np.linalg.matrix_power(y, i)

    P_Z = np.hstack((B.T, A.T))
    P_X = np.hstack((A, B))
    return CSScode(P_Z, P_X)


def generalised_bicycle_code(
    l: int, powers_A: list[int], powers_B: list[int]
) -> CSScode:
    """Construct the generalised bicycle code with circulant of size l,
    see eg https://arxiv.org/abs/1904.02703, https://arxiv.org/abs/1212.6703.

    Args:
        l: Size of circulant.
        powers_A: Powers of x representing first circulant matrix A.
        powers_B: Powers of xrepresenting second circulant matrix B.

    Return:
        The bivariate bicycle code.
    """
    identity_l = np.eye(l)

    A = np.zeros((l, l))
    for i in powers_A:
        A += np.roll(identity_l, i, axis=1)
    B = np.zeros((l, l))
    for i in powers_B:
        B += np.roll(identity_l, i, axis=1)

    P_Z = np.hstack((B.T, A.T))
    P_X = np.hstack((A, B))
    return CSScode(P_Z, P_X)


### Functions specific to bivariate bicycle codes ###


def polynomial_to_qubits(
    poly: list[tuple[int]], l: int, m: int, primed: bool = False
) -> list[int]:
    """Converts a polynomial in two variables into a set
    of qubits in a bivariate bicycle code.

    Args:
        poly: the polynomial in x and y to be converted.
        l: Size of first circulants in the tensor product.
        m: Size of second circulants in the tensor product.
        primed: Whether to index into the primed or
        unprimed block.

    Return:
        The set of indices i.e. qubits in the code.
    """
    indices = []
    for term in poly:
        index = term[0] * m + term[1]
        if primed:
            index += l * m
        indices.append(index)
    return indices


# See https://arxiv.org/abs/2308.07915 p23.
# alpha is a monomial in F_2[x, y], i.e. a pair of integers being powers of (x, y).
# f is a polynomial in F_2[x, y], i.e.
# a list of pairs of integers being powers of (x, y)
def unprimed_X_logical(
    alpha: tuple[int], f: list[tuple[int]], l: int, m: int
) -> list[int]:
    """Picks out a particular X logical in the unprimed block
    of a bivariate bicycle code, see p23 of
    https://arxiv.org/abs/2308.07915.

    Args:
        alpha: A monomial in F_2[x, y], i.e. a pair of integers
        being powers of x, y.
        f: A polynomial in F_2[x, y].
        l: Size of first circulants in the tensor product.
        m: Size of second circulants in the tensor product.

    Return:
        The set of indices i.e. qubits in the code, which the
        X logical has support on.
    """
    new_poly = []
    for mono in f:
        x_power = (alpha[0] + mono[0]) % l
        y_power = (alpha[1] + mono[1]) % m
        new_poly.append((x_power, y_power))
    return polynomial_to_qubits(new_poly, l, m, False)


def primed_X_logical(
    alpha: tuple[int], g: list[tuple[int]], h: list[tuple[int]], l: int, m: int
) -> list[int]:
    """Picks out a particular X logical in the primed block
    of a bivariate bicycle code, see p23 of
    https://arxiv.org/abs/2308.07915.

    Args:
        alpha: A monomial in F_2[x, y], i.e. a pair of integers
        being powers of x, y.
        g: A polynomial in F_2[x, y].
        h: A polynomial in F_2[x, y].
        l: Size of first circulants in the tensor product.
        m: Size of second circulants in the tensor product.

    Return:
        The set of indices i.e. qubits in the code, which the
        X logical has support on.
    """
    new_poly = []
    for mono in g:
        x_power = (alpha[0] + mono[0]) % l
        y_power = (alpha[1] + mono[1]) % m
        new_poly.append((x_power, y_power))
    indices1 = polynomial_to_qubits(new_poly, l, m, False)

    new_poly = []
    for mono in h:
        x_power = (alpha[0] + mono[0]) % l
        y_power = (alpha[1] + mono[1]) % m
        new_poly.append((x_power, y_power))
    indices2 = polynomial_to_qubits(new_poly, l, m, True)

    return indices1 + indices2


def unprimed_Z_logical(
    alpha: tuple[int], h: list[tuple[int]], g: list[tuple[int]], l: int, m: int
) -> list[int]:
    """Picks out a particular Z logical in the unprimed block
    of a bivariate bicycle code, see p23 of
    https://arxiv.org/abs/2308.07915.

    Args:
        alpha: A monomial in F_2[x, y], i.e. a pair of integers
        being powers of x, y.
        h: A polynomial in F_2[x, y].
        g: A polynomial in F_2[x, y].
        l: Size of first circulants in the tensor product.
        m: Size of second circulants in the tensor product.

    Return:
        The set of indices i.e. qubits in the code, which the
        Z logical has support on.
    """
    new_poly = []
    for mono in h:
        x_power = (alpha[0] - mono[0]) % l
        y_power = (alpha[1] - mono[1]) % m
        new_poly.append((x_power, y_power))
    indices1 = polynomial_to_qubits(new_poly, l, m, False)

    new_poly = []
    for mono in g:
        x_power = (alpha[0] - mono[0]) % l
        y_power = (alpha[1] - mono[1]) % m
        new_poly.append((x_power, y_power))
    indices2 = polynomial_to_qubits(new_poly, l, m, True)

    return indices1 + indices2


def primed_Z_logical(
    alpha: tuple[int], f: list[tuple[int]], l: int, m: int
) -> list[int]:
    """Picks out a particular Z logical in the primed block
    of a bivariate bicycle code, see p23 of
    https://arxiv.org/abs/2308.07915.

    Args:
        alpha: A monomial in F_2[x, y], i.e. a pair of integers
        being powers of x, y.
        f: A polynomial in F_2[x, y].
        l: Size of first circulants in the tensor product.
        m: Size of second circulants in the tensor product.

    Return:
        The set of indices i.e. qubits in the code, which the
        Z logical has support on.
    """
    new_poly = []
    for mono in f:
        x_power = (alpha[0] - mono[0]) % l
        y_power = (alpha[1] - mono[1]) % m
        new_poly.append((x_power, y_power))
    return polynomial_to_qubits(new_poly, l, m, True)



import copy
from dataclasses import dataclass

import numpy as np

from ssip.basic_functions import (
    CSScode,
    direct_sum_codes,
    find_paired_basis,
    find_X_basis,
    find_Z_basis,
    indices_to_vector,
    multiply_F2,
    num_data_qubits,
    num_logical_qubits,
    num_X_stabs,
    num_Z_stabs,
    vector_to_indices,
)
from ssip.coequalisers import internal_merge_by_indices
from ssip.lifted_product import tensor_product
from ssip.merge_result import MergeResult, find_logical_splitting
from ssip.monic_span_checker import (
    restricted_matrix,
)
from ssip.pushouts import pushout


@dataclass
class MergeSequence:
    """Data that informs `parallel_internal_merges` and
    `parallel_external_merges` which merges to perform.

    Attributes:
        IndexSequence: Each merge takes in a pair of sets of indices, corresponding to the logicals to be merged. An `IndexSequence` is hence a list of these pairs of sets.
        DepthSequence: A list of depths to perform the merges at.

    """

    IndexSequence: list
    DepthSequence: list


@dataclass
class MeasureSequence:
    """Data that informs `parallel_single_logical_qubit_measurements`
    which measurements to perform.

    Attributes:
        LogicalList: Each measurement requires a set of indices, i.e. the logical to be measured. A LogicalList is a list of these.
        DepthSequence: A list of depths to perform the merges at.

    """

    LogicalList: list
    DepthSequence: list


# see https://arxiv.org/abs/2110.10794.
# return_data primarily useful for obtaining the new Z and X logicals.
def measure_single_logical_qubit(
    C: CSScode,
    indices: list[int],
    basis: str = "Z",
    depth: int = 1,
    return_data: bool = False,
) -> CSScode | MergeResult:
    """Takes a CSS code and measures a logical qubit in the manner of
    https://arxiv.org/abs/2110.10794, in either the Z or X basis.

    Args:
        C: The CSScode.
        indices: The indices of the logical operator to be measured out.
        basis: The basis of the measurement.
        depth: Controls the size of the ancilla code used for measurement.
        return_data: Whether to return a MergeResult or a CSScode object.


    Returns:
        The merged CSScode, or the MergeResult object which contains the CSScode.
    """

    # generates a classical code which, when used in a tensor product,
    # can be used to make codes to measure single logical qubits.
    def generate_classical_line_code(n):
        mat = np.eye(n)
        for i in range(n - 1):
            mat[i + 1][i] = 1
        return mat

    restr, rows_to_keep = (None, None)

    if basis == "Z":
        restr, rows_to_keep = restricted_matrix(indices, C.PX)
    elif basis == "X":
        restr, rows_to_keep = restricted_matrix(indices, C.PZ)
    else:
        raise ValueError("Must enter a valid Pauli string.")

    rows_to_keep = sorted(rows_to_keep)

    V = restr
    S = generate_classical_line_code(depth)
    tens = tensor_product(S, V)
    n_initial = num_data_qubits(C)

    qmap = {v: k for k, v in enumerate(indices)}
    smap = {v: k for k, v in enumerate(rows_to_keep)}

    if basis == "Z":
        temp_code = pushout(
            C.PZ.T.copy(),
            C.PX.copy(),
            tens.PZ.T.copy(),
            tens.PX.copy(),
            qmap,
            smap,
            return_data,
        )
        if return_data:
            temp_code.MergeMap = temp_code.MergeMap[:, :n_initial]
            temp_code.NewZStabs = list(
                range(num_Z_stabs(C), num_Z_stabs(temp_code.Code))
            )
            temp_code.NewXStabs = list(
                range(num_X_stabs(C), num_X_stabs(temp_code.Code))
            )
            temp_code.NewQubits = list(
                range(num_data_qubits(C), num_data_qubits(temp_code.Code))
            )
        return temp_code
    elif basis == "X":
        temp_code = pushout(
            C.PX.T.copy(),
            C.PZ.copy(),
            tens.PZ.T.copy(),
            tens.PX.copy(),
            qmap,
            smap,
            return_data,
        )
        if return_data:
            temp_new_Z_logicals = temp_code.NewZLogicals.copy()
            temp_code.NewZLogicals = temp_code.NewXLogicals.copy()
            temp_code.NewXLogicals = temp_new_Z_logicals.copy()
            temp_old_Z_logicals = temp_code.OldZLogicals.copy()
            temp_code.OldZLogicals = temp_code.OldXLogicals.copy()
            temp_code.OldXLogicals = temp_old_Z_logicals.copy()
            temp_code.Code = CSScode(temp_code.Code.PX, temp_code.Code.PZ)
            temp_code.MergeMap = temp_code.MergeMap[:, :n_initial]
            temp_code.NewZStabs = list(
                range(num_Z_stabs(C), num_Z_stabs(temp_code.Code))
            )
            temp_code.NewXStabs = list(
                range(num_X_stabs(C), num_X_stabs(temp_code.Code))
            )
            temp_code.NewQubits = list(
                range(num_data_qubits(C), num_data_qubits(temp_code.Code))
            )

            return temp_code
        return CSScode(temp_code.PX, temp_code.PZ)


def parallel_single_logical_qubit_measurements(
    C: CSScode,
    measure_sequence: MeasureSequence,
    basis: str = "Z",
    return_data: bool = False,
) -> CSScode | MergeResult:
    """Takes a CSS code and measures several logical qubits in the manner of
    https://arxiv.org/abs/2110.10794, in either the Z or X basis, in parallel.

    Args:
        C: The CSScode.
        measure_sequence: The measurements to be performed.
        basis: The basis of the measurements.
        return_data: Whether to return a MergeResult or a CSScode object.


    Returns:
        The merged CSScode, or the MergeResult object which contains the CSScode.
    """
    merge_result = None
    total_merge_map = None

    num_measures = len(measure_sequence.LogicalList)
    for i in range(num_measures):
        if i == 0:
            merge_result = measure_single_logical_qubit(
                C,
                measure_sequence.LogicalList[0],
                basis,
                measure_sequence.DepthSequence[0],
                True,
            )
            if merge_result is None:
                raise RuntimeError(
                    "Parallel logical measurements cannot be found for the given input"
                )

            total_merge_map = merge_result.MergeMap
        else:
            n = total_merge_map.shape[1]
            v = indices_to_vector(measure_sequence.LogicalList[i], n)
            new_v = vector_to_indices(
                list((multiply_F2(total_merge_map, np.array([v]).T).T)[0])
            )

            merge_result = measure_single_logical_qubit(
                merge_result.Code,
                new_v,
                basis,
                measure_sequence.DepthSequence[i],
                True,
            )

            if merge_result is None:
                raise ValueError(
                    "Parallel logical measurements cannot be found for the given input"
                )
            total_merge_map = multiply_F2(merge_result.MergeMap, total_merge_map)

    if return_data:
        num_before_logicals = num_logical_qubits(C)
        num_after_logicals = num_logical_qubits(merge_result.Code)
        new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals = (
            find_logical_splitting(
                C,
                merge_result.Code,
                total_merge_map,
                num_before_logicals,
                num_after_logicals,
                num_measures,
                basis,
            )
        )
        new_Z_stabs = list(range(num_Z_stabs(C), num_Z_stabs(merge_result.Code)))
        new_X_stabs = list(range(num_X_stabs(C), num_X_stabs(merge_result.Code)))
        new_qubits = list(range(num_data_qubits(C), num_data_qubits(merge_result.Code)))

        return MergeResult(
            merge_result.Code,
            total_merge_map,
            new_Z_stabs,
            new_X_stabs,
            new_qubits,
            new_Z_logicals,
            new_X_logicals,
            old_Z_logicals,
            old_X_logicals,
        )

    return merge_result.Code


def parallel_internal_merges(
    C: CSScode,
    merge_sequence: MergeSequence,
    basis: str = "Z",
    return_data: bool = False,
) -> CSScode | MergeResult:
    """Takes a CSS code and performs several internal merges in parallel.

    Args:
        C: The CSScode.
        merge_sequence: The merges to be performed.
        basis: The basis of the measurements.
        return_data: Whether to return a MergeResult or a CSScode object.


    Returns:
        The merged CSScode, or the MergeResult object which contains the CSScode.
    """
    merge_result = None
    total_merge_map = None
    total_new_X_stabs = 0
    total_new_Z_stabs = 0
    num_new_qubits = 0
    num_merges = len(merge_sequence.IndexSequence)
    for i in range(num_merges):
        if i == 0:
            merge_result = internal_merge_by_indices(
                C,
                merge_sequence.IndexSequence[i][0],
                merge_sequence.IndexSequence[i][1],
                basis,
                merge_sequence.DepthSequence[i],
                True,
            )
            if merge_result is None:
                raise ValueError(
                    "parallel internal merges cannot be found for the given input"
                )

            total_merge_map = merge_result.MergeMap
            if merge_result.NewQubits is not None:
                num_new_qubits += len(merge_result.NewQubits)
            if merge_result.NewZStabs is not None:
                total_new_Z_stabs += len(merge_result.NewZStabs)
            if merge_result.NewXStabs is not None:
                total_new_X_stabs += len(merge_result.NewXStabs)

        else:
            n = total_merge_map.shape[1]
            v0 = indices_to_vector(merge_sequence.IndexSequence[i][0], n)
            v1 = indices_to_vector(merge_sequence.IndexSequence[i][1], n)

            new_v0 = vector_to_indices(
                list((multiply_F2(total_merge_map, np.array([v0]).T).T)[0])
            )
            new_v1 = vector_to_indices(
                list((multiply_F2(total_merge_map, np.array([v1]).T).T)[0])
            )

            merge_result = internal_merge_by_indices(
                merge_result.Code,
                new_v0,
                new_v1,
                basis,
                merge_sequence.DepthSequence[i],
                True,
            )
            if merge_result is None:
                raise ValueError(
                    "parallel internal merges cannot be found for the given input"
                )
            total_merge_map = multiply_F2(merge_result.MergeMap, total_merge_map)
            num_new_qubits += len(merge_result.NewQubits)
            total_new_Z_stabs += len(merge_result.NewZStabs)
            total_new_X_stabs += len(merge_result.NewXStabs)

    if return_data:
        new_Z_stabilisers = range(total_new_Z_stabs)
        new_X_stabilisers = range(total_new_X_stabs)

        num_before_logicals = num_logical_qubits(C)
        num_after_logicals = num_logical_qubits(merge_result.Code)
        num_new_logicals = num_after_logicals - num_before_logicals + num_merges
        new_qubits = [i + num_data_qubits(C) for i in range(num_new_qubits)]
        new_Z_logicals = []
        new_X_logicals = []
        old_Z_logicals = []
        old_X_logicals = []

        if num_new_logicals < 0:
            raise ValueError(
                "Internal merge error, logical operators are not irreducible"
            )
        if num_new_logicals:
            new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals = (
                find_logical_splitting(
                    C,
                    merge_result.Code,
                    total_merge_map,
                    num_before_logicals,
                    num_after_logicals,
                    num_merges,
                    basis,
                )
            )
        else:
            old_Z_logicals = find_Z_basis(merge_result.Code)
            old_X_logicals = find_X_basis(merge_result.Code)
            old_X_logicals = find_paired_basis(old_Z_logicals, old_X_logicals)

        return MergeResult(
            merge_result.Code,
            total_merge_map,
            new_Z_stabilisers,
            new_X_stabilisers,
            new_qubits,
            new_Z_logicals,
            new_X_logicals,
            old_Z_logicals,
            old_X_logicals,
        )

    return merge_result.Code


def parallel_external_merges(
    C: CSScode,
    D: CSScode,
    merge_sequence: MergeSequence,
    basis: str = "Z",
    return_data: bool = False,
) -> CSScode | MergeResult:
    """Takes two CSS codeblocks and performs several external merges in parallel.

    Args:
        C: The first CSS codeblock.
        D: The second CSS codeblock.
        merge_sequence: The merges to be performed.
        basis: The basis of the measurements.
        return_data: Whether to return a MergeResult or a CSScode object.


    Returns:
        The merged CSScode, or the MergeResult object which contains the CSScode.
    """
    merge_seq = copy.deepcopy(merge_sequence)
    sum_code = direct_sum_codes(C, D)
    n_C = num_data_qubits(C)
    for indices in merge_seq.IndexSequence:
        indices[1] = [i + n_C for i in indices[1]]
    return parallel_internal_merges(sum_code, merge_seq, basis, return_data)



import shutil
import subprocess
from ast import literal_eval
from functools import reduce
from pathlib import Path

import numpy as np

from ssip.basic_functions import (
    BinMatrix,
    CSScode,
    compose_to_zero,
    find_X_basis,
    find_Z_basis,
    num_data_qubits,
    systolic_distance,
)


def Z_distance(C: CSScode) -> int:
    """Finds the Z distance of a CSS code in a naive manner, by enumerating over
    all Z logicals.

    Args:
        C: The CSScode.


    Return:
        The Z distance of C.
    """
    return systolic_distance(C.PZ.T, C.PX)


def X_distance(C: CSScode) -> int:
    """Finds the X distance of a CSS code in a naive manner, by enumerating over
    all X logicals.

    Args:
        C: The CSScode.


    Return:
        The X distance of C.
    """
    return systolic_distance(C.PX.T, C.PZ)


def code_distance(C: CSScode, verbose: bool = False) -> int:
    """Finds the distance of a CSS code in a naive manner, by enumerating over
    all logicals.

    Args:
        C: The CSScode.


    Return:
        The distance of C.
    """
    dZ = Z_distance(C)
    if verbose:
        print(f"{dZ=}")
    dX = X_distance(C)
    if verbose:
        print(f"{dX=}")
    return min(dZ, dX)


# taken almost verbatim from Simon Burton's Qumba, see https://github.com/punkdit/qumba
def distance_lower_bound_z3(Hx: BinMatrix, Lx: BinMatrix, d: int) -> list | None:
    """Checks to see whether there are any Z logicals of weight at most d. If there are,
    returns the first one found.

    Args:
        Hx: The X parity check matrix of the code.
        Lx: A spanning set of X logicals of the code.
        d: The weight to check below.


    Return:
        The first nontrivial Z logical of weight at most d, if there is one.
        Otherwise returns None.
    """
    from z3 import Bool, If, Not, Or, Solver, Sum, Xor, sat

    m, n = Hx.shape
    k, n1 = Lx.shape

    if n != n1:
        raise ValueError("Stabiliser and logical matrices must have compatible sizes.")

    solver = Solver()
    add = solver.add
    v = [Bool("v%d" % i) for i in range(n)]

    term = Sum([If(v[i], 1, 0) for i in range(n)]) <= d
    add(term)

    def check(hx):
        terms = [v[j] for j in range(n) if hx[j]]
        if len(terms) > 1:
            return reduce(Xor, terms)
        elif len(terms) == 1:
            return terms[0]
        raise RuntimeError("dead check")

    # parity checks
    for i in range(m):
        add(Not(check(Hx[i])))

    # non-trivial logical
    term = reduce(Or, [check(Lx[i]) for i in range(k)])
    add(term)

    result = solver.check()
    if result != sat:
        return

    model = solver.model()
    v = [model.evaluate(v[i]) for i in range(n)]
    v = [int(literal_eval(str(vi))) for vi in v]

    assert compose_to_zero(Hx, np.array([v]).T), "bug bug... try updating z3?"
    assert not compose_to_zero(Lx, np.array([v]).T), "bug bug... try updating z3?"
    assert sum(v) <= d, "sum(v)=%d: bug bug... try updating z3?" % sum(v)
    return v


def distance_z3(
    C: CSScode, Lz: BinMatrix | None = None, Lx: BinMatrix | None = None
) -> int:
    """Finds the distance of a CSS code. Searches for logicals with weight 1, then
    weight 2 and so on. Works best for codes with low distances, but is implemented
    using Z3 so is quite fast.

    Args:
        C: The CSScode.
        Lz: A spanning set of Z logicals of the code.
        Lx: A spanning set of X logicals of the code.


    Return:
        The distance of C.
    """
    n = num_data_qubits(C)
    if Lz is None:
        Lz = np.array(find_Z_basis(C))
    if Lx is None:
        Lx = np.array(find_X_basis(C))

    d_X = 1
    while d_X < n:
        v = distance_lower_bound_z3(C.PZ, Lz, d_X)
        if v is not None:
            break
        d_X += 1

    d_Z = 1
    while d_Z < d_X:
        v = distance_lower_bound_z3(C.PX, Lx, d_Z)
        if v is not None:
            break
        d_Z += 1

    return min(d_Z, d_X)


# GAP QDistRnd I/O hack
def distance_GAP(
    C: CSScode,
    gap_path: str | None = None,
    num_information_sets: int | None = None,
    input_filename: str | None = None,
    output_filename: str | None = None,
) -> int:
    """Gives an upper bound on the distance of a CSS code using GAP. If the code is small
    enough and the number of information sets is high then this upper bound is tight.
    Empirically, we find that setting num_information_sets = 10,000 gives
    complete accuracy at ~200 qubits.

    Args:
        C: The CSScode.
        gap_path: The absolute path to the GAP directory.
        num_information_sets: The number of information sets to make.
        input_filename: Which file to use as input to GAP.
        output_filename: Which file to receive results from GAP.


    Return:
        An upper bound on the distance of C.
    """
    if input_filename is None:
        input_filename = "input.g"
    if output_filename is None:
        output_filename = "output.txt"
    if gap_path is None:
        gap_path = "gap"
    if num_information_sets is None:
        # empirically, 10,000 gives complete accuracy at ~200 qubits
        num_information_sets = 10000

    # Do the first way round...
    with Path.open(Path(input_filename), "w") as f:
        f.write('LoadPackage("QDistRnd");; \nF:=GF(2);;\n')
        str_PX = "Hx:=One(F)*["
        for i in range(C.PX.shape[0]):
            str_row = "["
            int_row = list(C.PX[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PX += str_row
        str_PX = str_PX[:-1] + "];;\n"
        f.write(str_PX)

        str_PZ = "Hz:=One(F)*["
        for i in range(C.PZ.shape[0]):
            str_row = "["
            int_row = list(C.PZ[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PZ += str_row
        str_PZ = str_PZ[:-1] + "];;\n"
        f.write(str_PZ)

        f.write(
            "d:= DistRandCSS(Hz,Hx,"
            + str(num_information_sets)
            + ',0,2 : field:=F);; \nPrintTo("'
            + output_filename
            + '", d);; \nquit;'
        )
        f.close()

    subprocess.run(
        [
            str(shutil.which(gap_path)),
            "--nointeract",
            "--norepl",
            input_filename,
        ],
        shell=False,  # noqa: S603
        stdout=subprocess.DEVNULL,
    )

    d1 = 0
    with Path.open(Path(output_filename)) as f:
        d1 = int(f.read())

    # Do the other way round...
    with Path.open(Path(input_filename), "w") as f:
        f.write('LoadPackage("QDistRnd");; \nF:=GF(2);;\n')
        str_PX = "Hx:=One(F)*["
        for i in range(C.PX.shape[0]):
            str_row = "["
            int_row = list(C.PX[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PX += str_row
        str_PX = str_PX[:-1] + "];;\n"
        f.write(str_PX)

        str_PZ = "Hz:=One(F)*["
        for i in range(C.PZ.shape[0]):
            str_row = "["
            int_row = list(C.PZ[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PZ += str_row
        str_PZ = str_PZ[:-1] + "];;\n"
        f.write(str_PZ)

        f.write(
            "d:= DistRandCSS(Hx,Hz,"
            + str(num_information_sets)
            + ',0,2 : field:=F);; \nPrintTo("'
            + output_filename
            + '", d);; \nquit;'
        )
        f.close()

    subprocess.run(
        [
            str(shutil.which(gap_path)),
            "--nointeract",
            "--norepl",
            input_filename,
        ],
        shell=False,  # noqa: S603
        stdout=subprocess.DEVNULL,
    )

    d2 = 0
    with Path.open(Path(output_filename)) as f:
        d2 = int(f.read())

    return min(d1, d2)


# Hack to use CSS min distance to find subsystem CSS dressed distance
def subsystem_distance_GAP(
    C: CSScode,
    gauge_Zs: list[list[int]],
    gauge_Xs: list[list[int]],
    gap_path: str | None = None,
    num_information_sets: int | None = None,
    input_filename: str | None = None,
    output_filename: str | None = None,
) -> int:
    """Gives an upper bound on the dressed distance of a subsystem CSS code using GAP.
    If the code is small enough and the number of information sets is high then this
    upper bound is tight.

    Args:
        C: The CSScode.
        gauge_Zs: A spanning set of the gauged Z logicals.
        gauge_Xs: A spanning set of the gauged X logicals.
        gap_path: The absolute path to the GAP directory.
        num_information_sets: The number of information sets to make.
        input_filename: Which file to use as input to GAP.
        output_filename: Which file to receive results from GAP.


    Return:
        An upper bound on the dressed distance of C as a subsystem code.
    """
    if input_filename is None:
        input_filename = "input.g"
    if output_filename is None:
        output_filename = "output.txt"
    if gap_path is None:
        gap_path = "gap"
    if num_information_sets is None:
        # empirically, 10,000 typically gives complete accuracy at ~200 qubits
        num_information_sets = 10000

    # Do the first way round...
    with Path.open(Path(input_filename), "w") as f:
        f.write('LoadPackage("QDistRnd");; \nF:=GF(2);;\n')
        str_PX = "Hx:=One(F)*["
        for i in range(C.PX.shape[0]):
            str_row = "["
            int_row = list(C.PX[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PX += str_row
        str_PX = str_PX[:-1] + "];;\n"
        f.write(str_PX)

        str_PZ = "Hz:=One(F)*["
        for i in range(C.PZ.shape[0]):
            str_row = "["
            int_row = list(C.PZ[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PZ += str_row
        for gauge_logical in gauge_Zs:
            str_row = "["
            for entry in gauge_logical:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PZ += str_row
        str_PZ = str_PZ[:-1] + "];;\n"
        f.write(str_PZ)

        f.write(
            "d:= DistRandCSS(Hz,Hx,"
            + str(num_information_sets)
            + ',0,2 : field:=F);; \nPrintTo("'
            + output_filename
            + '", d);; \nquit;'
        )
        f.close()

    subprocess.run(
        [
            str(shutil.which(gap_path)),
            "--nointeract",
            "--norepl",
            input_filename,
        ],
        shell=False,  # noqa: S603
        stdout=subprocess.DEVNULL,
    )

    d1 = 0
    with Path.open(Path(output_filename)) as f:
        d1 = int(f.read())

    # Do the other way round...
    with Path.open(Path(input_filename), "w") as f:
        f.write('LoadPackage("QDistRnd");; \nF:=GF(2);;\n')
        str_PX = "Hx:=One(F)*["
        for i in range(C.PX.shape[0]):
            str_row = "["
            int_row = list(C.PX[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PX += str_row
        for gauge_logical in gauge_Xs:
            str_row = "["
            for entry in gauge_logical:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PX += str_row
        str_PX = str_PX[:-1] + "];;\n"
        f.write(str_PX)

        str_PZ = "Hz:=One(F)*["
        for i in range(C.PZ.shape[0]):
            str_row = "["
            int_row = list(C.PZ[i])
            for entry in int_row:
                str_row += str(int(entry)) + ","
            str_row = str_row[:-1] + "],"
            str_PZ += str_row
        str_PZ = str_PZ[:-1] + "];;\n"
        f.write(str_PZ)

        f.write(
            "d:= DistRandCSS(Hx,Hz,"
            + str(num_information_sets)
            + ',0,2 : field:=F);; \nPrintTo("'
            + output_filename
            + '", d);; \nquit;'
        )
        f.close()

    subprocess.run(
        [
            str(shutil.which(gap_path)),
            "--nointeract",
            "--norepl",
            input_filename,
        ],
        shell=False,  # noqa: S603
        stdout=subprocess.DEVNULL,
    )

    d2 = 0
    with Path.open(Path(output_filename)) as f:
        d2 = int(f.read())

    return min(d1, d2)



# Functions for checking that two logical operators belonging to different codes do
# indeed form a basis-preserving irreducible monic span.
from collections.abc import Iterable

import networkx as nx
import numpy as np

from ssip.basic_functions import (
    BinMatrix,
    compose_to_zero,
    indices_to_vector,
    kernel_basis_calc,
)


# given a set of indices indicating the support of a vector in the domain
# calculates the matrix restricted to that support
def restricted_matrix(indices: list[int], B: BinMatrix) -> (BinMatrix, list[int]):
    width = len(indices)
    R = np.zeros((B.shape[0], width))
    # restrict columns based on indices in support
    for i in range(B.shape[0]):
        for j in range(width):
            R[i][j] = B[i][indices[j]]

    # restrict rows by removing all-zero rows
    rows_to_keep = []
    for i in range(R.shape[0]):
        all_zeros = True
        for j in range(width):
            if R[i][j]:
                all_zeros = False
                break
        if not all_zeros:
            rows_to_keep.append(i)

    height = len(rows_to_keep)
    R_2 = np.zeros((height, width))
    for i in range(height):
        for j in range(width):
            R_2[i][j] = R[rows_to_keep[i]][j]

    return (R_2, rows_to_keep)


# checks whether a logical contains any other logicals in its support
def is_irreducible_logical(indices: Iterable[int], B: BinMatrix) -> bool:
    """Takes a vector v in ker(B) and checks whether it is irreducible,
    i.e. has no subvectors in its support which are also in B. Does
    not check that the vector is a nontrivial logical, i.e. is not in
    im(A) for some check matrix A.

    Args:
        indices: The list of indices in the support of v.
        B: The check matrix for v.

    Returns:
        True if v is in ker(B) and there is no subvector of v which
        is also in ker(B). False otherwise.
    """
    vec = np.array([indices_to_vector(indices, B.shape[1])]).T
    if not compose_to_zero(B, vec):
        return False
    restr = restricted_matrix(indices, B)[0]
    return len(kernel_basis_calc(restr)) == 1


def monic_span(M: BinMatrix, N: BinMatrix) -> (dict, dict):
    """Checks that there is a monic span with a logical operator subcomplex at the apex,
    with each morphism
    mapping basis elements to basis elements, and returns the relevant data
    required to construct that span (there may be multiple,
    in which case it returns the first it finds).
    We approach this by checking that two boolean matrices M and N
    are equivalent up to permutation
    of rows/columns, i.e. PMQ = N for some permutation matrices P, Q.
    This is the same as finding a hypergraph isomorphism,
    if it exists, which can be reduced in poly-time
    (but at the cost of increased space) to finding
    a graph isomorphism for bipartite graphs, as described in p1 of
    "Colored Hypergraph Isomorphism is Fixed Parameter Tractable".
    We use NetworkX and the VF2 algorithm to find the graph isomorphisms.
    Despite formidable worst-case complexity, this routine is fast in practice.

    Args:
        M: The first F_2 matrix.
        N: The second F_2 matrix.

    Returns:
        The hypergraph isomorphism, which is a bijective map on columns and
        corresponding bijective map on rows.
    """
    num_vertices_M = M.shape[0]
    num_edges_M = M.shape[1]

    num_vertices_N = N.shape[0]
    num_edges_N = N.shape[1]

    if num_vertices_M != num_vertices_N:
        return None
    if num_edges_M != num_edges_N:
        return None

    num_vertices = num_vertices_M
    num_edges = num_edges_M
    GM = nx.Graph()
    # red node = was originally a vertex in the hypergraph
    # green node = was originally a hyperedge in the hypergraph
    for i in range(num_vertices):
        GM.add_nodes_from([(i, {"color": "red"})])
    for j in range(num_edges):
        GM.add_nodes_from([(j + num_vertices, {"color": "green"})])
    for i in range(num_vertices):
        for j in range(num_edges):
            if M[i][j]:
                GM.add_edge(i, j + num_vertices)

    GN = nx.Graph()
    for i in range(num_vertices):
        GN.add_nodes_from([(i, {"color": "red"})])
    for j in range(num_edges):
        GN.add_nodes_from([(j + num_vertices, {"color": "green"})])
    for i in range(num_vertices):
        for j in range(num_edges):
            if N[i][j]:
                GN.add_edge(i, j + num_vertices)

    iso = nx.vf2pp_isomorphism(GM, GN, node_label="color")
    if iso is None:
        return None
    qubit_dict = {}
    syndrome_dict = {}
    for i in iso:
        if i < num_vertices:
            syndrome_dict[i] = iso[i]
        else:
            qubit_dict[i - num_vertices] = iso[i] - num_vertices

    return (qubit_dict, syndrome_dict)



# Functions to perform "internal merges" on a code.
# Follows basically the same prescription as pushouts
# and is strictly more general.

# One can prove that these quotients on rows and columns are the correct ones
# using the quotient chain map.
import numpy as np

from ssip.basic_functions import (
    BinMatrix,
    CSScode,
    direct_sum_matrices,
    find_homology_basis,
    find_paired_basis,
    num_data_qubits,
    num_logical_qubits,
)
from ssip.merge_result import MergeResult, find_logical_splitting
from ssip.monic_span_checker import monic_span, restricted_matrix
from ssip.pushouts import sandwich_middle_code


# Assumes that there are no qubits which appear twice in the qubit_map;
# same for checks in the syndrome_map.
# A and B are differentials.
def coequaliser(
    A: BinMatrix,
    B: BinMatrix,
    qubit_map: dict,
    syndrome_map: dict,
    return_data: bool = False,
) -> CSScode | MergeResult:
    # First differential
    # Quotient rows together
    A2 = A.copy()
    B2 = B.copy()
    for item in qubit_map.items():
        # XOR, as two qubits being quotiented can share a Z-check
        A2[item[0]] = np.array(
            list(map(lambda x, y: bool((x + y) % 2), A2[item[0]], A2[item[1]]))
        )

    del2 = np.delete(A2, list(qubit_map.values()), 0)

    # Quotient rows together. Also an XOR
    for item in syndrome_map.items():
        B2[item[0]] = np.array(
            list(map(lambda x, y: bool((x + y) % 2), B2[item[0]], B2[item[1]]))
        )

    del1 = np.delete(B2, list(syndrome_map.values()), 0)

    # Second differential
    # Quotient columns together. Just an OR this time.
    for item in qubit_map.items():
        del1[:, item[0]] = np.array(
            list(map(lambda x, y: bool(x + y), del1[:, item[0]], del1[:, item[1]]))
        )

    del1 = np.delete(del1, list(qubit_map.values()), 1)
    new_code = CSScode(del2.T, del1)
    if return_data:
        coeq = np.eye(A.shape[0])
        for key, val in qubit_map.items():
            coeq[key] = np.array(
                list(map(lambda x, y: bool(x + y), coeq[key], coeq[val]))
            )
        coeq = np.delete(coeq, list(qubit_map.values()), 0)

        num_before_logicals = len(find_homology_basis(A, B))
        num_after_logicals = num_logical_qubits(new_code)
        num_new_logicals = num_after_logicals - num_before_logicals + 1

        new_Z_logicals = []
        new_X_logicals = []

        old_Z_logicals = []
        old_X_logicals = []

        if num_new_logicals < 0:
            raise ValueError("Coequaliser error, logical operators are not irreducible")
        if num_new_logicals:
            old_code = CSScode(A.T, B)
            new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals = (
                find_logical_splitting(
                    old_code, new_code, coeq, num_before_logicals, num_after_logicals
                )
            )
        else:
            old_Z_logicals = find_homology_basis(del2, del1)
            trial_X_basis = find_homology_basis(del1.T, del2.T)
            old_X_logicals = find_paired_basis(old_Z_logicals, trial_X_basis)

        return MergeResult(
            new_code,
            coeq,
            [],
            [],
            [],
            new_Z_logicals,
            new_X_logicals,
            old_Z_logicals,
            old_X_logicals,
        )

    return new_code


def coequaliser_by_indices(
    C: CSScode,
    indices1: list,
    indices2: list,
    basis: str = "Z",
    return_data: bool = False,
) -> CSScode | MergeResult | None:
    restr1 = None
    restr2 = None
    if basis == "Z":
        (restr1, rows_to_keep1) = restricted_matrix(indices1, C.PX)
        (restr2, rows_to_keep2) = restricted_matrix(indices2, C.PX)
    elif basis == "X":
        (restr1, rows_to_keep1) = restricted_matrix(indices1, C.PZ)
        (restr2, rows_to_keep2) = restricted_matrix(indices2, C.PZ)
    else:
        raise ValueError("Must enter a valid Pauli string.")
    span = monic_span(restr1, restr2)

    if span is None:
        return None
    qubit_map = {}
    for qb1, qb2 in span[0].items():
        qubit_map[indices1[qb1]] = indices2[qb2]
    syndrome_map = {}
    for syndrome1, syndrome2 in span[1].items():
        syndrome_map[rows_to_keep1[syndrome1]] = rows_to_keep2[syndrome2]

    if basis == "Z":
        return coequaliser(
            C.PZ.T.copy(), C.PX.copy(), qubit_map, syndrome_map, return_data
        )
    elif basis == "X":
        temp_code = coequaliser(
            C.PX.T.copy(), C.PZ.copy(), qubit_map, syndrome_map, return_data
        )
        if return_data:
            temp_new_Z_logicals = temp_code.NewZLogicals
            temp_code.NewZLogicals = temp_code.NewXLogicals
            temp_code.NewXLogicals = temp_new_Z_logicals
            temp_old_Z_logicals = temp_code.OldZLogicals
            temp_code.OldZLogicals = temp_code.OldXLogicals
            temp_code.OldXLogicals = temp_old_Z_logicals
            temp_code.Code = CSScode(temp_code.Code.PX, temp_code.Code.PZ)
            return temp_code
        return CSScode(temp_code.PX, temp_code.PZ)


# V the logical operator subcomplex, C the base code. Does not work if
# there are duplicate qubits or syndromes. The duplicates in the syndrome
# can be resolved mathematically if they appear in the same entry,
# but I do not know how to code this right now.
def internal_merge(
    V: BinMatrix,
    C: CSScode,
    qubit_map: dict,
    syndrome_map: dict,
    depth: int = 1,
    return_data: str = False,
) -> CSScode | MergeResult:
    tens = sandwich_middle_code(V, depth)
    direct_sum1 = direct_sum_matrices(tens.PZ.T, C.PZ.T)
    direct_sum2 = direct_sum_matrices(tens.PX, C.PX)

    num_new_qubits = num_data_qubits(tens)
    num_new_X_checks = tens.PX.shape[0]

    qmap1 = {}
    qmap2 = {}
    for count, item in enumerate(qubit_map.items()):
        qmap1[item[0] + num_new_qubits] = count
        qmap2[item[1] + num_new_qubits - len(qubit_map)] = (depth - 1) * len(
            qubit_map
        ) + count

    smap1 = {}
    smap2 = {}
    for count, item in enumerate(syndrome_map.items()):
        smap1[item[0] + num_new_X_checks] = count
        smap2[item[1] + num_new_X_checks - len(syndrome_map)] = (depth - 1) * len(
            syndrome_map
        ) + count

    first_coeq = coequaliser(direct_sum1, direct_sum2, qmap1, smap1)
    second_coeq = coequaliser(first_coeq.PZ.T, first_coeq.PX, qmap2, smap2)

    if return_data:
        n_before = num_data_qubits(C)
        n_after = num_data_qubits(second_coeq)

        if not n_after == num_data_qubits(C) + (depth - 1) * len(
            qubit_map
        ) + depth * len(syndrome_map):
            raise RuntimeError("Dimensions do not match up when returning merge data")

        new_qubits = list(
            range((depth - 1) * len(qubit_map) + depth * len(syndrome_map))
        )

        merge_map = np.zeros((n_after, n_before))
        for i in range(n_before):
            merge_map[i + len(new_qubits)][i] = 1

        new_Z_stabilisers = list(range(depth * len(qubit_map)))
        new_X_stabilisers = list(range((depth - 1) * len(syndrome_map)))

        num_before_logicals = num_logical_qubits(C)
        num_after_logicals = num_logical_qubits(second_coeq)
        num_new_logicals = num_after_logicals - num_before_logicals + 1

        new_Z_logicals = []
        new_X_logicals = []

        old_Z_logicals = []
        old_X_logicals = []

        if num_new_logicals < 0:
            raise ValueError(
                "Internal merge error, logical operators are not irreducible"
            )
        if num_new_logicals:
            new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals = (
                find_logical_splitting(
                    C, second_coeq, merge_map, num_before_logicals, num_after_logicals
                )
            )
        else:
            old_Z_logicals = find_homology_basis(second_coeq.PZ.T, second_coeq.PX)
            old_X_logicals = find_homology_basis(second_coeq.PX.T, second_coeq.PZ)
            old_X_logicals = find_paired_basis(old_Z_logicals, old_X_logicals)

        return MergeResult(
            second_coeq,
            merge_map,
            new_Z_stabilisers,
            new_X_stabilisers,
            new_qubits,
            new_Z_logicals,
            new_X_logicals,
            old_Z_logicals,
            old_X_logicals,
        )

    return second_coeq


def internal_merge_by_indices(
    C: CSScode,
    indices1: list[int],
    indices2: list[int],
    basis: str = "Z",
    depth: int = 1,
    return_data: bool = False,
) -> CSScode | MergeResult | None:
    """Construct an internal merge within one CSS codeblock, along two
    disjoint logical operators.

    Args:
        C: The CSScode to merge within.
        indices1: The qubits in the support of the first logical operator.
        indices2: The qubits in the support of the second logical operator.
        basis: The basis, X or Z, to perform the logical measurement in.
        depth: The depth of the merge, i.e. how large to make the tensor product intermediate code.
        return_data: Whether to calculate a MergeResult rather than a CSScode output. The MergeResult includes substantially more data about the merge.


    Return:
        The merged CSScode, or the MergeResult object which contains the CSScode, or None if no merge can be found.
    """
    restr1 = None
    restr2 = None
    if basis == "Z":
        (restr1, rows_to_keep1) = restricted_matrix(indices1, C.PX)
        (restr2, rows_to_keep2) = restricted_matrix(indices2, C.PX)
    elif basis == "X":
        (restr1, rows_to_keep1) = restricted_matrix(indices1, C.PZ)
        (restr2, rows_to_keep2) = restricted_matrix(indices2, C.PZ)
    else:
        raise ValueError("Must enter a valid Pauli string.")
    span = monic_span(restr1, restr2)
    if span is None:
        return None
    qubit_map = {}
    for qb1, qb2 in span[0].items():
        qubit_map[indices1[qb1]] = indices2[qb2]
    syndrome_map = {}
    for syndrome1, syndrome2 in span[1].items():
        syndrome_map[rows_to_keep1[syndrome1]] = rows_to_keep2[syndrome2]
    qubit_map = dict(sorted(qubit_map.items()))
    syndrome_map = dict(sorted(syndrome_map.items()))

    for val in qubit_map.values():
        if val in qubit_map:
            return None
    # if there are duplicate syndromes do not merge, as it yields a
    # qubit with no X-checks, hence the code is distance 1
    for val in syndrome_map.values():
        if val in syndrome_map:
            return None

    if basis == "Z":
        return internal_merge(restr1, C, qubit_map, syndrome_map, depth, return_data)
    elif basis == "X":
        temp_code = internal_merge(
            restr1, CSScode(C.PX, C.PZ), qubit_map, syndrome_map, depth, return_data
        )
        if return_data:
            temp_code.Code = CSScode(temp_code.Code.PX, temp_code.Code.PZ)
            temp_new_X_stabs = temp_code.NewXStabs.copy()
            temp_code.NewXStabs = temp_code.NewZStabs.copy()
            temp_code.NewZStabs = temp_new_X_stabs
            temp_new_Z_logicals = temp_code.NewZLogicals
            temp_code.NewZLogicals = temp_code.NewXLogicals
            temp_code.NewXLogicals = temp_new_Z_logicals
            temp_old_Z_logicals = temp_code.OldZLogicals
            temp_code.OldZLogicals = temp_code.OldXLogicals
            temp_code.OldXLogicals = temp_old_Z_logicals
            return temp_code
        return CSScode(temp_code.PX, temp_code.PZ)



# Functions for performing elementary calculations with CSS codes,
# relying on the CSS code-homology correspondence.

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from itertools import combinations

import numpy as np
from numpy.typing import NDArray

BinMatrix = NDArray[np.uint8]


@dataclass
class CSScode:
    """A qubit Calderbank-Shor-Steane code with a set of stabiliser generators
    is a pair of parity-check matrices over the field F_2.

    Attributes:
        PZ: Parity-check matrix over F_2 for Z generators.
        PX: Parity-check matrix over F_2 for X generators.
    """

    PZ: BinMatrix
    PX: BinMatrix


def flip_code(C: CSScode) -> CSScode:
    """Takes a CSS code and exchanges PX and PZ, giving a code with the same
    parameters but flipped Z and X generators.

    Args:
        C: The CSS code.

    Returns:
        A new CSS code with flipped X and Z generators.
    """
    return CSScode(C.PX, C.PZ)


# add a tuple of vectors over F_2 together
def vec_addition(vecs: Sequence[Sequence[int]]) -> list[int]:
    sum_vec = list(vecs[0])
    for j in range(1, len(vecs)):
        for i in range(len(sum_vec)):
            sum_vec[i] = int(sum_vec[i] != vecs[j][i])
    return sum_vec


def vector_to_indices(v: Sequence[int]) -> Sequence[int]:
    """Converts a vector over F_2 to a list of entries that vector
    has support on.

    Args:
        v: The vector to convert.

    Return:
        The list of indices that vector has non-zero support on.
    """
    indices = [i for i in range(len(v)) if v[i]]
    return indices


def indices_to_vector(indices: Iterable[int], length: int) -> list[int]:
    """Converts a list of indices to a vector with support on those
    indices.

    Args:
        indices: The list of indices.
        length: The desired length of the vector.

    Return:
        The vector with support on those indices.
    """
    v = [0] * length
    for i in indices:
        v[i] = 1
    return v


def dot_product(vec1: BinMatrix, vec2: BinMatrix) -> int:
    """Takes the bilinear symmetric nondegenerate form, i.e. dot product,
    F_2^n x F_2^n -> F_2.

    Args:
        vec1: The first F_2 vector.
        vec2: The second F_2 vector.

    Return:
        The dot product.
    """
    if not len(vec1) == len(vec2):
        raise ValueError("Incompatible lengths of vectors supplied to dot product")
    prod = 0
    for i in range(len(vec1)):
        prod += vec1[i] and vec2[i]
    return prod % 2


def multiply_F2(A: BinMatrix, B: BinMatrix) -> BinMatrix:
    """Multiplies two matrices over F_2.

    Args:
        A: First matrix.
        B: Second matrix.

    Return:
        The matrix AB.
    """
    mult = np.matmul(A, B)
    for i in range(mult.shape[0]):
        for j in range(mult.shape[1]):
            mult[i][j] %= 2
    return mult


def compose_to_zero(A: BinMatrix, B: BinMatrix) -> bool:
    """Check whether two matrices multiply to give zero.

    Args:
        A: First matrix.
        B: Second matrix.

    Return:
        True if AB = 0, False otherwise.
    """
    mult = np.matmul(A, B)
    for row in mult:
        for i in row:
            if int(i) % 2:
                return False
    return True


def is_valid_CSScode(C: CSScode) -> bool:
    """Check whether a CSScode is valid, i.e. the generators commute.

    Args:
        C: The CSScode to check.

    Return:
        True if the generators commute, False otherwise.
    """
    return compose_to_zero(C.PX, C.PZ.T)


def max_measurement_weight(C: CSScode):
    """Finds the maximum weight of any generator in a CSScode.

    Args:
        C: The CSScode to check.

    Return:
        The maximum number of qubits in the support of any generator.
    """

    wX = max(np.sum(C.PX, axis=1))
    wZ = max(np.sum(C.PZ, axis=1))
    return max(wX, wZ)


def max_measurement_support(C: CSScode) -> int:
    """Finds the maximum number of generators any single qubit in a
    CSScode is in the support of.

    Args:
        C: The CSScode to check.

    Return:
        The maximum number of generators any single qubit is in the support of.
    """
    sX = max(np.sum(C.PX, axis=0))
    sZ = max(np.sum(C.PZ, axis=0))
    return max(sX, sZ)


def num_data_qubits(C: CSScode) -> int:
    """Finds the number of data qubits in a CSScode.

    Args:
        C: The CSScode.

    Return:
        The number of data qubits.
    """
    return C.PZ.shape[1]


def num_X_stabs(C: CSScode) -> int:
    """Finds the number of X-type stabiliser generators in a CSScode.

    Args:
        C: The CSScode.

    Return:
        The number of X-type stabiliser generators.
    """
    return C.PX.shape[0]


def num_Z_stabs(C: CSScode) -> int:
    """Finds the number of Z-type stabiliser generators in a CSScode.

    Args:
        C: The CSScode.

    Return:
        The number of Z-type stabiliser generators.
    """
    return C.PZ.shape[0]


# constructs the augmented matrix and applies Gaussian elimination to acquire reduced
# column echelon form
def column_echelon_form(A: BinMatrix, augmented: bool = True):
    def swap_cols(mat: BinMatrix, col1: int, col2: int) -> None:
        if col1 != col2:
            for i in np.arange(mat.shape[0]):
                temp = mat[i][col1]
                mat[i][col1] = mat[i][col2]
                mat[i][col2] = temp

    # add column 1 to column 2, stored in column 2
    def col_addition(mat, col1, col2):
        for i in np.arange(mat.shape[0]):
            mat[i][col2] = mat[i][col1] != mat[i][col2]

    # make augmented matrix
    B = np.array(0)
    if augmented:
        B = np.zeros((A.shape[0] + A.shape[1], A.shape[1]))
        for i in np.arange(A.shape[1]):
            B[A.shape[0] + i][i] = 1
    else:
        B = np.zeros((A.shape[0], A.shape[1]))

    for i in np.arange(A.shape[0]):
        for j in np.arange(A.shape[1]):
            B[i][j] = A[i][j]
    # Gaussian elimination
    h = 0
    k = 0
    while h < B.shape[1] and k < A.shape[0]:
        pivot_exists = False
        pivot_col = 0
        for i in np.arange(h, B.shape[1]):
            if B[k][i]:
                pivot_exists = True
                pivot_col = i
        if not pivot_exists:
            k += 1

        else:
            swap_cols(B, h, pivot_col)
            for i in np.arange(0, B.shape[1]):
                if B[k][i] and i != h:
                    col_addition(B, h, i)
            k += 1
            h += 1
    return B


def kernel_basis_calc(A: BinMatrix) -> list[list[int]]:
    """Calculates a basis of the kernel of a matrix.

    Args:
        A: The matrix.

    Return:
        A basis of the kernel of the matrix.
    """
    B = column_echelon_form(A, True)
    basis = []
    for j in np.arange(B.shape[1] - 1, -1, -1):
        all_zeros = True
        for i in np.arange(A.shape[0]):
            if B[i][j]:
                all_zeros = False
                break
        if not all_zeros:
            break
        basis_vec = [0] * (B.shape[1])
        for i in np.arange(A.shape[0], B.shape[0]):
            basis_vec[i - A.shape[0]] = B[i][j]
        basis.append(basis_vec)

    return basis


def kernel_all_vectors(A: BinMatrix) -> list[list[int]]:
    """Calculates all nonzero vectors in the kernel of a matrix.

    Args:
        A: The matrix.

    Return:
        All nonzero vectors in the kernel of A.
    """
    basis = kernel_basis_calc(A)
    all_combs = []
    for i in range(1, len(basis) + 1):
        all_combs += list(combinations(basis, i))
    for i in range(len(all_combs)):
        all_combs[i] = vec_addition(all_combs[i])

    return all_combs


def image_basis_calc(A: BinMatrix) -> list[list[int]]:
    """Calculates a basis of the image of a matrix.

    Args:
        A: The matrix.

    Return:
        A basis of the image of the matrix.
    """
    B = column_echelon_form(A, False)
    basis = []
    for j in np.arange(B.shape[1]):
        all_zeros = True
        for i in np.arange(B.shape[0]):
            if B[i][j]:
                all_zeros = False
        if all_zeros:
            break
        else:
            basis_vec = [0] * (B.shape[0])
            for i in np.arange(B.shape[0]):
                basis_vec[i] = B[i][j]
            basis.append(basis_vec)

    return basis


def image_all_vectors(A: BinMatrix) -> list[list[int]]:
    """Calculates all nonzero vectors in the image of a matrix.

    Args:
        A: The matrix.

    Return:
        All nonzero vectors in the image of A.
    """
    basis = image_basis_calc(A)
    all_combs = []
    for i in range(1, len(basis) + 1):
        all_combs += list(combinations(basis, i))
    for i in range(len(all_combs)):
        all_combs[i] = vec_addition(all_combs[i])

    return all_combs


def num_logical_qubits(C: CSScode) -> int:
    """Calculate the number of logical qubits of a CSScode.

    Args:
        C: The CSScode.

    Return:
        The number of logical qubits of C.
    """
    im = image_basis_calc(C.PZ.T)
    ker = kernel_basis_calc(C.PX)
    return len(ker) - len(im)


# Finds a basis for the homology space, i.e. the quotient ker(B)/im(A).
# Works by first finding the quotient morphism Q: F_2^n -> F_2^n/im(A),
# then performing Gaussian elimination to calculate the linearly independent
# vectors in ker(B) under the equivalence /imA.
def find_homology_basis(A: BinMatrix, B: BinMatrix) -> list[list[int]]:
    im_basis = np.array(image_basis_calc(A))
    im_dim = im_basis.shape[0]
    quotient_map = np.array(kernel_basis_calc(im_basis))

    ker_basis = np.array(kernel_basis_calc(B)).T
    ker_dim = ker_basis.shape[1]

    hom_dim = ker_dim - im_dim
    if hom_dim < 0:
        raise ValueError("Kernel is smaller than image")

    quotiented_vecs = multiply_F2(quotient_map, ker_basis)
    height = quotiented_vecs.shape[0]
    mat = column_echelon_form(quotiented_vecs, True)

    vecs_to_add = [0] * hom_dim
    for i in np.arange(hom_dim):
        v = [0] * (mat.shape[0] - height)
        for j in np.arange(height, mat.shape[0]):
            v[j - height] = mat[j][i]
        vecs_to_add[i] = v

    hom_basis = [0] * hom_dim
    for i in np.arange(hom_dim):
        vecs = [
            list(ker_basis[:, j])
            for j in np.arange(len(vecs_to_add[i]))
            if vecs_to_add[i][j]
        ]

        hom_basis[i] = vec_addition(vecs)

    return hom_basis


def find_Z_basis(C: CSScode) -> list[list[int]]:
    """Calculate a basis of Z logical operators of a CSScode.

    Args:
        C: The CSScode.

    Return:
        A basis of Z logical operators of C.
    """
    return find_homology_basis(C.PZ.T, C.PX)


def find_X_basis(C: CSScode) -> list[list[int]]:
    """Calculate a basis of X logical operators of a CSScode.

    Args:
        C: The CSScode.

    Return:
        A basis of X logical operators of C.
    """
    return find_homology_basis(C.PX.T, C.PZ)


def find_paired_basis(
    basis1: list[list[int]], basis2: list[list[int]]
) -> list[list[int]]:
    """Calculate a paired basis of logical operators of a CSScode. That is,
    given two bases of the homology H_1 and cohomology H^1 respectively,
    find and return a basis of H^1 which is 'paired' with that of H_1 by the
    nondegenerate bilinear form, i.e. the dot product.

    Args:
        basis1: The basis of H_1.
        basis2: The basis of H^1.

    Return:
        A basis of H^1 which is paired with basis1.
    """
    if len(basis1) != len(basis2):
        raise ValueError(
            "Bases have lengths " + str(len(basis1)) + " & " + str(len(basis2))
        )
    m = len(basis1)

    mat = np.zeros((m, m), dtype=np.uint8)
    for i in range(m):
        for j in range(m):
            mat[i][j] = dot_product(basis1[i], basis2[j])
    B = column_echelon_form(mat, True)
    replacement_basis = []
    for j in np.arange(B.shape[1]):
        vecs_to_add = [
            basis2[i - B.shape[1]]
            for i in np.arange(mat.shape[0], B.shape[0])
            if B[i][j]
        ]

        sum_vec = vec_addition(vecs_to_add)
        replacement_basis.append(sum_vec)

    return replacement_basis


# given a (co)chain complex C2 -> C1 -> C0, with A: C2 -> C1 and B: C1 -> C0
# returns the systolic distance at degree 1
def systolic_distance(A: BinMatrix, B: BinMatrix) -> int:
    quotient_map = np.array(kernel_basis_calc(A.T))
    ker_vecs = kernel_all_vectors(B)
    sys_distance = len(ker_vecs[0])
    for v in ker_vecs:
        weight = int(sum(v))
        if weight < sys_distance and not compose_to_zero(quotient_map, np.array([v]).T):
            sys_distance = weight
    return sys_distance


def direct_sum_matrices(A: BinMatrix, B: BinMatrix) -> BinMatrix:
    """Compute the direct sum of two matrices.

    Args:
        A: The top left matrix.
        B: The bottom right matrix.

    Return:
        The direct sum of A and B.
    """
    rows = A.shape[0] + B.shape[0]
    cols = A.shape[1] + B.shape[1]
    direct_sumAB = np.zeros((rows, cols), dtype=np.uint8)
    direct_sumAB[: A.shape[0], : A.shape[1]] = A
    direct_sumAB[A.shape[0] :, A.shape[1] :] = B
    return direct_sumAB


def direct_sum_codes(C: CSScode, D: CSScode) -> CSScode:
    """Compute the direct sum of two codes (as chain or cochain complexes).

    Args:
        C: The first code.
        D: The second code.

    Return:
        The direct sum of C and D.
    """
    return CSScode(direct_sum_matrices(C.PZ, D.PZ), direct_sum_matrices(C.PX, D.PX))



# Functions for constructing the two pushouts necessary
# for a fault-tolerant code merge between CSS codes.
import numpy as np

from ssip.basic_functions import (
    BinMatrix,
    CSScode,
    direct_sum_matrices,
    find_homology_basis,
    find_paired_basis,
    num_data_qubits,
    num_logical_qubits,
)
from ssip.lifted_product import tensor_product
from ssip.merge_result import MergeResult, find_logical_splitting
from ssip.monic_span_checker import monic_span, restricted_matrix


# Calculates the tensor product in the category of chain complexes
# of a logical operator subcomplex V and the complex P1 -> P0,
# with delP: P1 -> P0 = (1 1)^T.
# See p32 of https://arxiv.org/abs/2301.13738.
def sandwich_middle_code(V: BinMatrix, depth: int = 1) -> CSScode:
    def gen_classical_code(depth):
        parity_mat = np.zeros((depth + 1, depth))
        for i in range(depth):
            parity_mat[i][i] = 1
            parity_mat[i + 1][i] = 1
        return parity_mat

    A = gen_classical_code(depth)
    return tensor_product(A, V)


# Calculates the pushout of a basis-preserving monic span with
# logical operator subcomplex as
# the apex. Assumes we have A, B differentials belonging
# to one complex and C, D to another.
def pushout(
    A: BinMatrix,
    B: BinMatrix,
    C: BinMatrix,
    D: BinMatrix,
    qubit_map: dict,
    syndrome_map: dict,
    return_data: bool = False,
) -> CSScode | MergeResult:
    # the first differential
    direct_sumAC = direct_sum_matrices(A, C)

    # quotient rows together
    for item in qubit_map.items():
        direct_sumAC[item[0]] = np.array(
            list(
                map(
                    lambda x, y: bool(x + y),
                    direct_sumAC[item[0]],
                    direct_sumAC[item[1] + A.shape[0]],
                )
            )
        )

    del2 = np.delete(direct_sumAC, [x + A.shape[0] for x in qubit_map.values()], 0)
    # the second differential
    direct_sumBD = direct_sum_matrices(B, D)

    # quotient rows together
    for item in syndrome_map.items():
        direct_sumBD[item[0]] = np.array(
            list(
                map(
                    lambda x, y: bool(x + y),
                    direct_sumBD[item[0]],
                    direct_sumBD[item[1] + B.shape[0]],
                )
            )
        )

    del1 = np.delete(direct_sumBD, [x + B.shape[0] for x in syndrome_map.values()], 0)

    # quotient columns together
    for item in qubit_map.items():
        del1[:, item[0]] = np.array(
            list(
                map(
                    lambda x, y: bool(x + y),
                    del1[:, item[0]],
                    del1[:, item[1] + B.shape[1]],
                )
            )
        )
    del1 = np.delete(del1, [x + B.shape[1] for x in qubit_map.values()], 1)

    new_code = CSScode(del2.T, del1)

    if return_data:
        # Calculate coequaliser at degree 1
        n = A.shape[0] + C.shape[0]
        n2 = A.shape[0]
        coeq = np.eye(n)
        for key, val in qubit_map.items():
            coeq[key] = np.array(
                list(map(lambda x, y: bool(x + y), coeq[key], coeq[val + n2]))
            )
        coeq = np.delete(coeq, [x + n2 for x in qubit_map.values()], 0)

        # Calculate whether there are new logical qubits introduced by the merge
        num_before_logicals = len(find_homology_basis(A, B)) + len(
            find_homology_basis(C, D)
        )
        num_after_logicals = num_logical_qubits(new_code)
        num_new_logicals = num_after_logicals - num_before_logicals + 1

        new_Z_logicals = []
        new_X_logicals = []

        old_Z_logicals = []
        old_X_logicals = []

        if num_new_logicals < 0:
            raise ValueError("Pushout error, logical operators are not irreducible")
        # Calculate paired Z/X logicals
        if num_new_logicals:
            old_code = CSScode(direct_sum_matrices(A, C).T, direct_sum_matrices(B, D))
            new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals = (
                find_logical_splitting(
                    old_code, new_code, coeq, num_before_logicals, num_after_logicals
                )
            )

        else:
            old_Z_logicals = find_homology_basis(del2, del1)
            trial_X_basis = find_homology_basis(del1.T, del2.T)
            old_X_logicals = find_paired_basis(old_Z_logicals, trial_X_basis)

        return MergeResult(
            new_code,
            coeq,
            [],
            [],
            [],
            new_Z_logicals,
            new_X_logicals,
            old_Z_logicals,
            old_X_logicals,
        )

    return new_code


# Calculates the pushout based on qubits labelled by indices; if these indices do not
# yield a monic span then return None. Assume C, D are CSScodes.
def pushout_by_indices(
    C: CSScode,
    D: CSScode,
    indices1: list[int],
    indices2: list[int],
    basis: str = "Z",
    return_data: bool = False,
) -> CSScode | MergeResult | None:
    restr1 = None
    restr2 = None
    if basis == "Z":
        restr1 = restricted_matrix(indices1, C.PX)
        restr2 = restricted_matrix(indices2, D.PX)
    elif basis == "X":
        restr1 = restricted_matrix(indices1, C.PZ)
        restr2 = restricted_matrix(indices2, D.PZ)
    else:
        raise ValueError("Must enter a valid Pauli string.")

    span = monic_span(restr1[0], restr2[0])
    if span is None:
        return None
    qubit_map = {}
    for qb1, qb2 in span[0].items():
        qubit_map[indices1[qb1]] = indices2[qb2]
    syndrome_map = {}
    for syndrome1, syndrome2 in span[1].items():
        syndrome_map[restr1[1][syndrome1]] = restr2[1][syndrome2]
    if basis == "Z":
        return pushout(
            C.PZ.T.copy(),
            C.PX.copy(),
            D.PZ.T.copy(),
            D.PX.copy(),
            qubit_map,
            syndrome_map,
            return_data,
        )
    elif basis == "X":
        temp_code = pushout(
            C.PX.T.copy(),
            C.PZ.copy(),
            D.PX.T.copy(),
            D.PZ.copy(),
            qubit_map,
            syndrome_map,
            return_data,
        )
        if return_data:
            temp_new_Z_logicals = temp_code.NewZLogicals.copy()
            temp_code.NewZLogicals = temp_code.NewXLogicals.copy()
            temp_code.NewXLogicals = temp_new_Z_logicals.copy()
            temp_old_Z_logicals = temp_code.OldZLogicals.copy()
            temp_code.OldZLogicals = temp_code.OldXLogicals.copy()
            temp_code.OldXLogicals = temp_old_Z_logicals.copy()
            temp_code.Code = CSScode(temp_code.Code.PX, temp_code.Code.PZ)
            return temp_code
        return CSScode(temp_code.PX, temp_code.PZ)


# Given two CSS codes C2 -> C1 -> C0 and D2 -> D1 -> C0, with F: C2 -> C1, G: C1 -> 0,
# J: D2 -> D1, K: D1 -> D0, and matching logical operators u in C1 and v in D1,
# constructs the pushout of a pushout of chain complexes, giving the final code.
# Assumes that the pushouts are of basis-preserving monic spans with logical operator
# subcomplexes at the apexes.
def external_merge(
    V: BinMatrix,
    F: BinMatrix,
    G: BinMatrix,
    J: BinMatrix,
    K: BinMatrix,
    qubit_map: dict,
    syndrome_map: dict,
    depth: int = 1,
    return_data: str = False,
) -> CSScode | MergeResult:
    if depth < 1:
        raise ValueError("Must enter a pushout depth greater than 0.")

    tens = sandwich_middle_code(V, depth)
    qmap1 = {}
    qmap2 = {}
    for count, item in enumerate(qubit_map.items()):
        qmap1[count] = item[0]
        qmap2[count + depth * len(qubit_map)] = item[1]
    smap1 = {}
    smap2 = {}
    for count, item in enumerate(syndrome_map.items()):
        smap1[count] = item[0]
        smap2[count + depth * len(syndrome_map)] = item[1]

    first_pushout = pushout(tens.PZ.T, tens.PX, F, G, qmap1, smap1)
    second_pushout = pushout(first_pushout.PZ.T, first_pushout.PX, J, K, qmap2, smap2)

    if return_data:
        n_before = G.shape[1] + K.shape[1]
        n_after = num_data_qubits(second_pushout)
        if not n_after == G.shape[1] + K.shape[1] + (depth - 1) * len(
            qubit_map
        ) + depth * len(syndrome_map):
            raise RuntimeError("Dimensions do not match up when returning merge data")

        new_qubits = list(np.arange(len(qubit_map), depth * len(qubit_map))) + list(
            np.arange(
                (depth + 1) * len(qubit_map),
                (depth + 1) * len(qubit_map) + depth * len(syndrome_map),
            )
        )
        new_Z_stabilisers = list(range(depth * len(qubit_map)))
        new_X_stabilisers = list(range(len(syndrome_map), depth * len(syndrome_map)))

        merge_map = np.zeros((n_after, n_before))

        qmap1_flipped = {v: k for k, v in qmap1.items()}
        qmap2_flipped = {v: k for k, v in qmap2.items()}

        count1 = 0
        count2 = 0

        for i in range(G.shape[1]):
            if i in qmap1_flipped:
                merge_map[qmap1_flipped[i]][i] = 1
                count1 += 1
            else:
                merge_map[
                    (depth + 1) * len(qubit_map)
                    + depth * len(syndrome_map)
                    + i
                    - count1
                ][i] = 1

        for i in range(K.shape[1]):
            if i in qmap2_flipped:
                merge_map[qmap2_flipped[i]][i + G.shape[1]] = 1
                count2 += 1
            else:
                j = (
                    G.shape[1]
                    + depth * len(qmap2)
                    + depth * len(syndrome_map)
                    + i
                    - count2
                )
                merge_map[j][i + G.shape[1]] = 1

        num_before_logicals = num_logical_qubits(CSScode(F.T, G)) + num_logical_qubits(
            CSScode(J.T, K)
        )
        num_after_logicals = num_logical_qubits(second_pushout)
        num_new_logicals = num_after_logicals - num_before_logicals + 1
        new_Z_logicals = []
        new_X_logicals = []

        old_Z_logicals = []
        old_X_logicals = []

        if num_new_logicals < 0:
            raise ValueError(
                "External merge error, logical operators are not irreducible"
            )
        if num_new_logicals:
            old_code = CSScode(direct_sum_matrices(F, J).T, direct_sum_matrices(G, K))
            new_Z_logicals, new_X_logicals, old_Z_logicals, old_X_logicals = (
                find_logical_splitting(
                    old_code,
                    second_pushout,
                    merge_map,
                    num_before_logicals,
                    num_after_logicals,
                )
            )
        else:
            old_Z_logicals = find_homology_basis(second_pushout.PZ.T, second_pushout.PX)
            trial_X_logicals = find_homology_basis(
                second_pushout.PX.T, second_pushout.PZ
            )
            old_X_logicals = find_paired_basis(old_Z_logicals, trial_X_logicals)

        return MergeResult(
            second_pushout,
            merge_map,
            new_Z_stabilisers,
            new_X_stabilisers,
            new_qubits,
            new_Z_logicals,
            new_X_logicals,
            old_Z_logicals,
            old_X_logicals,
        )

    return second_pushout


def external_merge_by_indices(
    C: CSScode,
    D: CSScode,
    indices1: list[int],
    indices2: list[int],
    basis: str = "Z",
    depth: int = 1,
    return_data: bool = False,
) -> CSScode | MergeResult | None:
    """Construct an external merge between two CSS codeblocks, along two
    logical operators.

    Args:
        C: The first CSS codeblock.
        D: The second CSS codeblock.
        indices1: The qubits in the support of the first logical operator.
        indices2: The qubits in the support of the second logical operator.
        basis: The basis, X or Z, to perform the logical measurement in.
        depth: The depth of the merge, i.e. how large to make the tensor product intermediate code.
        return_data: Whether to calculate a MergeResult rather than a CSScode output. The MergeResult includes substantially more data about the merge.


    Return:
        The merged CSScode, or the MergeResult object which contains the CSScode, or None if no merge could be found.
    """
    restr1 = None
    restr2 = None

    if basis == "Z":
        restr1 = restricted_matrix(indices1, C.PX)
        restr2 = restricted_matrix(indices2, D.PX)
    elif basis == "X":
        restr1 = restricted_matrix(indices1, C.PZ)
        restr2 = restricted_matrix(indices2, D.PZ)
    else:
        raise ValueError("Must enter a valid Pauli string.")

    span = monic_span(restr1[0], restr2[0])
    if span is None:
        return None

    qubit_map = {}
    for qb1, qb2 in span[0].items():
        qubit_map[indices1[qb1]] = indices2[qb2]
    syndrome_map = {}
    for syndrome1, syndrome2 in span[1].items():
        syndrome_map[restr1[1][syndrome1]] = restr2[1][syndrome2]
    qubit_map = dict(sorted(qubit_map.items()))
    syndrome_map = dict(sorted(syndrome_map.items()))
    if basis == "Z":
        return external_merge(
            restr1[0],
            C.PZ.T.copy(),
            C.PX.copy(),
            D.PZ.T.copy(),
            D.PX.copy(),
            qubit_map,
            syndrome_map,
            depth,
            return_data,
        )
    if basis == "X":
        temp_code = external_merge(
            restr1[0],
            C.PX.T.copy(),
            C.PZ.copy(),
            D.PX.T.copy(),
            D.PZ.copy(),
            qubit_map,
            syndrome_map,
            depth,
            return_data,
        )
        if return_data:
            temp_code.Code = CSScode(temp_code.Code.PX, temp_code.Code.PZ)
            temp_new_X_stabs = temp_code.NewXStabs.copy()
            temp_code.NewXStabs = temp_code.NewZStabs.copy()
            temp_code.NewZStabs = temp_new_X_stabs
            temp_new_Z_logicals = temp_code.NewZLogicals.copy()
            temp_code.NewZLogicals = temp_code.NewXLogicals.copy()
            temp_code.NewXLogicals = temp_new_Z_logicals.copy()
            temp_old_Z_logicals = temp_code.OldZLogicals.copy()
            temp_code.OldZLogicals = temp_code.OldXLogicals.copy()
            temp_code.OldXLogicals = temp_old_Z_logicals.copy()
            return temp_code
        return CSScode(temp_code.PX, temp_code.PZ)
